# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@163.com
Date: 2021-11-5 15:17
Desc:
'''

import sys
import math
import inspect
import numpy as np

# 1.进度条展示 progress_bar(1, 100)
def progress_bar(portion, total):
    """
    total 总数据大小，portion 已经传送的数据大小
    :param portion: 已经接收的数据量
    :param total: 总数据量
    :return: 接收数据完成，返回True
    """
    part = total / 50  # 1%数据的大小
    count = math.ceil(portion / part)
    sys.stdout.write('\r')
    sys.stdout.write(('[%-50s]%.2f%%' % (('>' * count), portion / total * 100)))
    sys.stdout.flush()

    if portion >= total:
        sys.stdout.write('\n')
        return True
    return False

# 2.获取变量名
def get_retrieve_name(var):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    list_name = [var_name for var_name, var_val in callers_local_vars if var_val is var]
    return list_name

# 3.计算加权方差
def get_weighted_std(values,weights):
    average = np.average(values, weights=weights)
    variance = np.average((values-average)**2, weights=weights)
    return variance

# 4.计算等频分箱
def get_frequence_blocks(x, box_num=10, drop_duplicates=True):
    '''
        等频分箱
    :param x: dtype=[]
    :param bins: 分箱数量
    :return: 箱边界
    '''
    x=np.sort(x).astype(float)
    len_clocks = int(min(len(x),box_num))
    ind = np.linspace(0, len(x)-1, len_clocks+1).astype(int)
    blocks=np.array(x)[ind].astype(float)
    if drop_duplicates:
        blocks = np.unique(blocks)
    if len(blocks)<2:
        return [-np.inf,np.inf]
    blocks[0] = -np.inf
    blocks[-1] = np.inf
    return blocks.tolist()

# 4.计算等距分箱
def get_distince_blocks(x, box_num=5, drop_duplicates=True):
    '''
        等距分箱
    :param x: dtype=[]
    :param bins: 分箱数量
    :return: 箱边界
    '''
    blocks = np.linspace(min(x), max(x), box_num+1).astype(float)
    if drop_duplicates:
        blocks = np.unique(blocks)
    if len(blocks)<2:
        return [-np.inf,np.inf]
    blocks[0] = -np.inf
    blocks[-1] = np.inf
    return blocks.tolist()


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 12:23
Desc:
'''

import numpy as np

def get_randomforest_blocks(clf_rf, col_name=None):
    '''
        利用随机森林获得最优分箱的边界值列表
    '''
    rf_boundary = []
    for i, treei in enumerate(clf_rf.estimators_):
        n_nodes = treei.tree_.node_count
        feature = treei.tree_.feature
        if col_name is not None:
            feature = [(col_name[i] if i>-1 else i)for i in feature]
        threshold = treei.tree_.threshold
        children_left = treei.tree_.children_left
        children_right = treei.tree_.children_right

        tree_boundary = {}
        for i in range(n_nodes):
            if children_left[i] != children_right[i]:  # 获得决策树节点上的划分边界值(除叶子结点)
                if feature[i] in tree_boundary:
                    boundary = tree_boundary.get(feature[i])
                    boundary.append(threshold[i])
                else:
                    boundary = [-np.inf, np.inf]
                    boundary.append(threshold[i])
                tree_boundary.update({feature[i]: boundary})
        for k, v in tree_boundary.items():
            sv = list(set([round(i, 4) for i in v]))
            sv.sort()
            tree_boundary.update({k: sv})

        rf_boundary.append(tree_boundary)
    return rf_boundary



# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 21:51
Desc:
'''
from interval import Interval
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from score_ecard.features import randomforest_blocks as rf_bolcks
from score_ecard.features import layered_woe as woe
from util import progress_bar, get_weighted_std
import pandas as pd
import numpy as np
import warnings
from score_ecard import score_card
warnings.filterwarnings('ignore')

class ECardModel():
    def __init__(self,kwargs_lr=None, kwargs_rf=None):
        self.params_rf = {
            'n_estimators': 100,
            'max_depth': 10,
            'max_features': 0.8,
            'min_samples_leaf': 0.05,
            'bootstrap': True,
            'random_state': 666
        }
        self.params_lr = {
            "class_weight": 'balanced',
            "penalty": 'l1',
            "C": 0.2,
            "solver": 'liblinear'
        }
        if kwargs_rf:
            self.params_rf.update(**kwargs_rf)
        if kwargs_lr:
            self.params_lr.update(**kwargs_lr)
        self.rf_cards = []
        self.score_model = score_card.ScoreCardModel()

    def fit(self, df_X, df_Y, validation_X:pd.DataFrame=None, validation_Y:pd.DataFrame=None, sample_weight=None):
        df_y, df_Y = self.check_label(df_Y)
        clf_rf = RandomForestClassifier(**self.params_rf)
        clf_rf.fit(df_X, df_y, sample_weight=sample_weight)
        rf_boundaries = rf_bolcks.get_randomforest_blocks(clf_rf,col_name=df_X.columns)
        gl_boundaries = self.get_boundaries_merge(rf_boundaries)
        self.blocks = gl_boundaries
        init_ecard = self.get_init_ecard(df_X,gl_boundaries)
        print("start model fitting ...")
        pre_y=np.array([0])

        # vlidation part
        validation_idx = False
        if type(validation_X) != type(None):
            assert validation_X.shape[0] == validation_Y.shape[0]
            df_valy, df_valY = self.check_label(validation_Y)
            pre_valy = np.array([0])
            validation_idx=True

        for i,tree_bins in enumerate(rf_boundaries):
            clf_lr = LogisticRegression(**self.params_lr)
            sample_idx = df_X.index.isin(df_X.sample(frac=0.7, weights=sample_weight, random_state=i).index)
            df_woe = woe.get_woe_card(df_X[sample_idx], df_Y[sample_idx], tree_bins)
            x = self.get_woe_features(df_X, df_woe, tree_bins)
            clf_lr.fit(x[sample_idx], df_y[sample_idx], sample_weight=sample_weight)
            clf_lr.col_name=x.columns
            tree_card = self.get_score_card(clf_lr,df_woe)
            self.rf_cards.append(tree_card)
            pre_y = pre_y + clf_lr.predict_proba(x)[:,1]
            cur_pre = (pre_y/(i+1))
            train_auc = self.score_model.get_auc(cur_pre,df_y, pre_target=1)[0]
            validation_info=None
            if validation_idx:
                valx = self.get_woe_features(validation_X, df_woe, tree_bins)
                pre_valy = pre_valy + clf_lr.predict_proba(valx)[:, 1]
                cur_pre = (pre_valy / (i + 1))
                val_auc = self.score_model.get_auc(cur_pre, df_valy, pre_target=1)[0]
                validation_info = "vlidation_auc={}".format(val_auc)
            print("sep_{}:\tauc={} {}".format(i+1,train_auc,validation_info))
        self.score_ecard = self.get_cards_merge(self.rf_cards, init_ecard)

    def check_label(self, df_Y):
        if df_Y.shape[-1] == 1:
            df_Y.columns = ['label']
        if 'label' not in df_Y.columns:
            df_Y['label'] = df_Y.iloc[:,0]
        df_y=df_Y.label
        return df_y, df_Y

    def get_woe_features(self, df_X,df_card,dict_blocks):
        df_x = pd.DataFrame()
        len_ = len(dict_blocks)
        for i, (_field, _bin) in enumerate(dict_blocks.items()):
            progress_bar(i, len_-1)
            _card = df_card[df_card.特征字段 == _field]
            _card.loc[:,'分箱'] = _card['分箱'].astype(str)
            _data = pd.cut(df_X[_field].fillna(0), bins=_bin)
            out = pd.DataFrame(_data).astype(str).join(_card.set_index('分箱'), on=_data.name, how='left')
            col = [i for i in out.columns if 'woe' in i]
            col2 = [_field + "_" + i for i in col]
            df_x[col2] = out[col]
        return df_x

    def get_score_card(self,clf_lr,df_woe):
        coef_dict = dict(zip(clf_lr.col_name, clf_lr.coef_[0]))
        init_score = self.score_model.score_offset
        base_score = self.score_model.score_factor * clf_lr.intercept_[0]
        woe_col = [i for i in df_woe.columns if 'woe' in i]
        get_coef_summary = lambda x: np.sum([coef_dict.get(x['特征字段'] + '_' + i) * x[i] for i in woe_col])
        df_woe['model_score'] = self.score_model.score_factor * df_woe.apply(get_coef_summary, axis=1)
        df_field_card = df_woe[['特征字段', '分箱', '车辆总数', 'model_score']]
        df_field_card.columns = ['field_', 'bins_', 'size_', 'score_']
        df_base_card = pd.DataFrame(
            [['init_base_score', Interval(-np.inf, np.inf, closed='right'), None, init_score],
             ['init_model_score', Interval(-np.inf, np.inf, closed='right'), None, base_score]
             ], columns=['field_', 'bins_', 'size_', 'score_'])
        df_card = df_field_card.append(df_base_card, ignore_index=True)
        return df_card

    def get_cards_merge(self, rf_cards, init_ecard):
        df_ecard=init_ecard.copy()
        df_ecard['score_']=0
        df_ecard['bins_']=df_ecard['bins_'].astype(str)
        df_ecard = df_ecard.groupby(['field_','bins_','size_'])['score_'].sum()
        len_ = len(rf_cards)
        for i,tree_card in enumerate(rf_cards):
            progress_bar(i,len_)
            icard = init_ecard.join(tree_card.set_index('field_')[['bins_', 'score_']], on='field_', how='left',
                                    rsuffix='_')
            idx = icard.apply(
                lambda x: True if str(x['bins__']) != 'nan' and (x['bins_'].overlaps(x['bins__'])) else False, axis=1)
            icard.loc[~idx, 'score_'] = 0
            icard['bins_']=icard['bins_'].astype(str)
            df_icard = icard.groupby(['field_', 'bins_', 'size_'])['score_'].sum()
            df_ecard+=df_icard/len_
        return df_ecard.reset_index()

    def get_boundaries_merge(self, boundaries_list):
        gl_boundaries={}
        for boundaries in boundaries_list:
            for k,v in boundaries.items():
                if k in gl_boundaries:
                    lv = gl_boundaries.get(k)
                    cv = list(set(v+lv))
                    cv.sort()
                    gl_boundaries.update({k:cv})
                else:
                    gl_boundaries.update({k: v})
        print("get randomforest boundaries done")
        return gl_boundaries

    def get_init_ecard(self, df_X, gl_boundaries):
        len_=len(gl_boundaries)
        print("init ecards ...")
        bins_list=[]
        for i, (col, bins) in enumerate(gl_boundaries.items()):
            progress_bar(i,len_-1)
            data_bin = pd.cut(df_X.loc[:, col].fillna(0), bins=bins).value_counts()
            df_bin = pd.DataFrame(data_bin).sort_index().reset_index()
            df_bin.columns = ['bins_','size_']
            df_bin.insert(loc=0, column='field_', value=col)
            bins_list.append(df_bin)
        df_init_ecard = pd.concat(bins_list).reset_index(drop=True)
        df_base_card = pd.DataFrame(
            [['init_base_score', Interval(-np.inf, np.inf, closed='right'), -1],
             ['init_model_score', Interval(-np.inf, np.inf, closed='right'), -1]
             ], columns=['field_', 'bins_', 'size_'])
        df_card = df_init_ecard.append(df_base_card, ignore_index=True)
        print("init ecards done")
        return df_card

    def get_importance_(self):
        df_card_info = self.score_ecard.groupby("field_").agg(
            {'size_': [("weights", lambda x: list(x))],
             'score_': [("values", lambda x: list(x))]

             }).droplevel(0, 1)
        importance_ = df_card_info.apply(lambda x: get_weighted_std(x["values"], x["weights"]), axis=1)
        return importance_.sort_values(ascending=False)

    def predict_proba(self, df_data):
        df_score = self.predict(df_data)
        df_proba =df_score.apply(lambda x:self.score_model.score_to_probability(x))
        return df_proba

    def predict(self, df_data):
        df_score = self.get_batch_score(df_data)
        return df_score

    # 单车详细评分
    def get_single_score(self, data: dict, level_threshold=None):
        score = 0
        score_detail = {}
        df_card = self.score_ecard
        for i, row in df_card.iterrows():
            field_ = row['field_']
            bins_ = row['bins_']
            score_ = row['score_']
            class_ = row['class_']
            if field_ in ['init_base_score', 'init_model_score']:
                v = 0
            else:
                v = data.get(field_)
            try:
                v = float(v)
            except:
                score = 0
                score_detail[field_] = '异常'
                break
            if not (v > -np.inf):
                v = 0
            if v in bins_:
                score = score + score_
                score_detail[field_] = class_
            else:
                pass

        # level
        level = -1
        total_meters = data.get('run_meters', 0)
        total_seconds = data.get('run_seconds', 0)
        if (level_threshold is not None) and (total_meters >= 2000000) and (total_seconds >= 10 * 3600):
            level = np.argwhere(np.sort(level_threshold + [score]) == score)[0][0]

        out = {'score': round(score), 'level': level, 'score_detail': score_detail}
        return out

    def get_batch_score(self, df_data: pd.DataFrame):
        '''
        ## 批量查询
        :param df_data:
        :return:
        '''
        df_score = pd.DataFrame()
        df_card = self.score_ecard
        for field_, bin_ in self.blocks.items():
            _card = df_card[df_card.field_ == field_]
            _card["bins_"] = _card.bins_.astype(str)
            _data = pd.cut(df_data[field_].fillna(0), bins=bin_).astype(str)
            out = pd.DataFrame(_data).join(_card.set_index('bins_'), on=_data.name, how='left')
            df_score[field_] = out['score_']
        df_score['init_base_score'] = df_card[df_card.field_ == 'init_base_score']['score_'].values[0]
        df_score['init_model_score'] = df_card[df_card.field_ == 'init_model_score']['score_'].values[0]
        score = df_score.sum(axis=1).apply(lambda x: round(x))
        return score


if __name__ == '__main__':
    df_valid = pd.read_csv("data/train_test_data.csv")
    df_train_data = df_valid[df_valid['train_test_tag'] == '训练集']
    df_test_data = df_valid[df_valid['train_test_tag'] == '测试集']
    features_col = ['trip_avg_meters',
                    'trip_avg_seconds', 'trip_avg_distance', 'high_meters_ratio',
                    'province_meters_ratio', 'high_trip_cnt_ratio',
                    'province_trip_cnt_ratio', 'curvature_g2_trip_meters_ratio',
                    'ng_23_6_seconds_ratio', 'ng_23_6_trip_cnt_ratio', 'daily_run_kmeters',
                    'daily_run_hours', 'daily_trip_cnt', 'daily_nohigh_kmeters',
                    'daily_ng_23_6_hours', 'trip_long_cnt_ratio', 'day_high_meters_ratio',
                    'ng_province_meters_ratio', 'morn_6_10_seconds_ratio',
                    'dusk_17_20_seconds_ratio', 'ng_23_6_avg_speed', 'morn_6_10_avg_speed',
                    'dusk_17_20_avg_speed', 'low_speed_seconds_ratio',
                    'low_speed_block_cnt_ratio', 'week_1_5_seconds_ratio',
                    'geohash4_top10_meters_ratio', 'trip_r30m_cnt_ratio',
                    'common_line_top30_cnt_ratio', 'ratio_meitan', 'ratio_gangtie', 'ratio_shashi',
                    'ratio_kuaidi', 'ratio_nonglinmufu', 'ratio_jiadian', 'ratio_jixie',
                    'ratio_qiche', 'ratio_fengdian', 'ratio_other'
                    ]
    df_X = df_train_data[features_col]
    df_Y = df_train_data[['label', 'label_ordinary',
                          'label_serious', 'label_major', 'label_devastating', 'label_8w']]
    ecard = ECardModel(kwargs_rf={'n_estimators':2})
    ecard.fit(df_X, df_Y,df_X, df_Y)
    print(ecard.get_importance_())



# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/5 22:08
Desc:
'''

import numpy as np
import pandas as pd
from util import progress_bar


class WeightOfEvidence:
    def __init__(self,bad_target=1):
        self.bad_target = bad_target
        self.eps = 1e-4
        self.woe_card = None

    def fit(self,df_x,df_y):
        '''
        :param df_x: 特征字段
        :param df_y: 标签分类字段
        :return:
            返回woe详细信息
        '''
        self.cur_field = df_x.name or 'field'
        # |fields |bins |bad |good |bad_prob |good_prob |woe |iv
        bin = np.unique(df_x)
        good = np.array([np.logical_and(df_x == val, df_y != self.bad_target).sum() for val in bin])
        bad = np.array([np.logical_and(df_x == val, df_y == self.bad_target).sum() for val in bin])
        prob_good = good / float(np.sum(df_y != self.bad_target))
        prob_bad = bad / float(np.sum(df_y==self.bad_target))
        woe = np.log(np.maximum(prob_good,self.eps)/np.maximum(prob_bad,self.eps))
        iv = (prob_good-prob_bad)*woe
        df_info = pd.DataFrame(data=np.array([bin,good,bad,prob_good,prob_bad,woe,iv]).T,
                               columns=['bin','good','bad','good_prob','bad_prob','woe','iv'])
        df_info.insert(0, 'fields', self.cur_field)
        self.woe_card=df_info

    def transform(self,x):
        card = self.woe_card
        if type(x) in (pd.DataFrame,pd.Series):
            df_x = pd.DataFrame(x)
            df_x.columns = ['bin']
            out = pd.DataFrame(df_x).join(card.set_index('bin'), on='bin', how='left')
        else:
            out = card[card.bin == x].woe
        return out

    def fit_transform(self,df_x,df_y):
        self.fit(df_x,df_y)
        df_x = pd.DataFrame(df_x)
        df_x.columns=['bin']
        out = pd.DataFrame(df_x).join(self.woe_card.set_index('bin'),on='bin',how='left')
        return out


class WeightOfFeeEvidence:
    def __init__(self,bad_target=1):
        self.bad_target = bad_target
        self.eps = 1e-4
        self.woe_card = None

    def fit(self,df_x,df_fee_got,df_report_fee):
        '''
        :param df_x: 特征字段
        :param df_fee_got: 已赚保费
        :param df_report_fee: 赔付金额
        :return:
            返回woe详细信息
        '''
        self.cur_field = df_x.name or 'field'
        # |fields |bins |bad |good |bad_prob |good_prob |woe |iv
        bin = np.unique(df_x)
        good = np.array([((df_x == val)*df_fee_got).sum() for val in bin])
        bad = np.array([((df_x == val)*df_report_fee).sum() for val in bin])
        prob_good = good / float(np.sum(df_fee_got))
        prob_bad = bad / float(np.sum(df_report_fee))
        woe = np.log(np.maximum(prob_good,self.eps)/np.maximum(prob_bad,self.eps))
        iv = (prob_good-prob_bad)*woe
        df_info = pd.DataFrame(data=np.array([bin,good,bad,prob_good,prob_bad,woe,iv]).T,
                               columns=['bin','get_fee','report_fee','get_fee_prob','report_fee_prob','woe','iv'])
        df_info.insert(0, 'fields', self.cur_field)
        self.woe_card=df_info

    def transform(self,x):
        card = self.woe_card
        if type(x) in (pd.DataFrame,pd.Series):
            df_x = pd.DataFrame(x)
            df_x.columns = ['bin']
            out = pd.DataFrame(df_x).join(card.set_index('bin'), on='bin', how='left')
        else:
            out = card[card.bin == x].woe
        return out

    def fit_transform(self, df_x, df_fee_got, df_report_fee):
        self.fit(df_x, df_fee_got, df_report_fee)
        df_x = pd.DataFrame(df_x)
        df_x.columns = ['bin']
        out = pd.DataFrame(df_x).join(self.woe_card.set_index('bin'), on='bin', how='left')
        return out

def get_woe_card(df_X, df_Y, fields_bins):
    '''

    :param df_X:
    :param df_Y:
    :param fields_bins: 特征的分箱字典
    :return: 构建woe字典
    '''
    assert 'label' in df_Y.columns
    # assert 'fee_got' in df_Y.columns
    # assert 'report_fee' in df_Y.columns

    woe_list = []
    len_=len(fields_bins)
    for i,(col, bins) in enumerate(fields_bins.items()):
        progress_bar(i,len_-1)
        bin_data = pd.cut(df_X.loc[:,col].fillna(0), bins=bins)

        # label1：是否出险
        woe1 = WeightOfEvidence()
        woe1.fit(bin_data, df_Y["label"])
        card = woe1.woe_card
        card.columns = ['特征字段', '分箱', '无出险车辆数', '出险车辆数',
                        '无出险占比', '出险占比', 'woe1', 'iv1']
        card['车辆总数'] = card['无出险车辆数'] + card['出险车辆数']

        # label2：赔付率
        if ('fee_got' in df_Y.columns) and ('report_fee' in df_Y.columns):
            woe2 = WeightOfFeeEvidence()
            woe2.fit(bin_data, df_Y["fee_got"], df_Y["report_fee"])
            card2 = woe2.woe_card
            card[["已赚保费", "赔付金额", "已赚保费占比", "赔付金额占比", "woe0", "iv0"]] = card2[
                ['get_fee', 'report_fee', 'get_fee_prob', 'report_fee_prob', 'woe', 'iv']]

        # label3：是否有过赔付2k+
        if 'label_ordinary' in df_Y.columns:
            woe3 = WeightOfEvidence()
            woe3.fit(bin_data, df_Y["label_ordinary"])
            card3 = woe3.woe_card
            card[["赔付车辆数", "无赔付车辆数", "赔付车辆占比", "无赔付车辆占比", "woe2", "iv2"]] = card3[
                ['good', 'bad', 'good_prob', 'bad_prob', 'woe', 'iv']]

        # label4：是否大事故1w+
        if 'label_serious' in df_Y.columns:
            woe4 = WeightOfEvidence()
            woe4.fit(bin_data, df_Y["label_serious"])
            card4 = woe4.woe_card
            card[["重大事故车辆数", "非重大事故车辆数", "重大事故占比", "非重大事故占比", "woe3", "iv3"]] = card4[
                ['good', 'bad', 'good_prob', 'bad_prob', 'woe', 'iv']]

        # label5：是否重大事故5w+
        if 'label_major' in df_Y.columns:
            woe5 = WeightOfEvidence()
            woe5.fit(bin_data, df_Y["label_major"])
            card5 = woe5.woe_card
            card[["特大事故车辆数", "非特大事故车辆数", "特大事故占比", "非特大事故占比", "woe4", "iv4"]] = card5[
                ['good', 'bad', 'good_prob', 'bad_prob', 'woe', 'iv']]

        # label6：是否特大事故20W'+
        if 'label_devastating' in df_Y.columns:
            woe5 = WeightOfEvidence()
            woe5.fit(bin_data, df_Y["label_devastating"])
            card5 = woe5.woe_card
            card[["特大事故车辆数", "非特大事故车辆数", "特大事故占比", "非特大事故占比", "woe5", "iv5"]] = card5[
                ['good', 'bad', 'good_prob', 'bad_prob', 'woe', 'iv']]

        woe_list.append(card)
    woe_dict = pd.concat(woe_list).reset_index(drop=True)
    woe_dict.columns.name = 'idx'
    col = ['特征字段', '分箱', '车辆总数'] + [i for i in woe_dict if 'woe' in i]+ [i for i in woe_dict if 'iv' in i]
    df_woe = woe_dict[col]
    return df_woe


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/11/6 00:11
Desc:
'''


import numpy as np
import pandas as pd

## 评分体系
class ScoreCardModel:
    def __init__(self):
        self.load_score_alpha()

    def load_score_alpha(self, odds=1 / 2, bScore=600, addScore=-30):
        '''
            bScore = offset - factor*ln(odds)
            bScore + addScore = offset - factor*ln(2*odds)
        :param odds: 坏好比
        :param bScore: 基础score分
        :param addScore: 评分变动值
        :return: bScore = offset - factor*ln(odds)
        '''

        self.score_factor = addScore / np.log(2)
        self.score_offset = bScore - addScore * np.log(odds) / np.log(2)

    def probability_to_score(self, bad_prob, eps=1e-5):
        '''
            逾期概率转换为评分
        :param bad_prob: 逾期概率
        :return: bScore = offset - factor*ln(odds)
        '''

        odds = max(bad_prob, eps) / max(1 - bad_prob, eps)
        score = self.score_offset + self.score_factor * np.log(odds)
        return int(round(score))

    def score_to_probability(self, score, eps=1e-5):
        '''
            评分转换为逾期概率
        :param score: 信用评分
        :return: bScore = offset - factor*ln(odds) -->bad_prob
        '''
        score_factor = eps if self.score_factor==0 else self.score_factor
        odds = np.e**((score - self.score_offset)/(score_factor))
        odds = -1+eps if odds == -1 else odds
        bad_prob = 1.00*odds/(1+odds)
        return bad_prob

    @classmethod
    def get_auc(cls, df_pre, df_label, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签{0,1}
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        crossfreq = pd.crosstab(df_pre, df_label).sort_index(ascending=False)  # 按照预测概率从大到小排序作为阈值
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        auc = np.trapz(y=tpr, x=fpr)
        return auc, crossdens

    @classmethod
    def get_g7_auc(cls, df_pre, df_got_fee, df_report_fee):
        '''
        功能: 计算auc值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_got_fee: 一维数组或series，代表已赚保费
        :param df_report_fee: 一维数组或series，代表赔付额
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        df_data = pd.concat(
            [df_pre.reset_index(drop=True), df_got_fee.reset_index(drop=True), df_report_fee.reset_index(drop=True)],
            axis=1, ignore_index=True)
        df_data.columns = ['pre', 'got_fee', 'report_fee']
        df_data = df_data.set_index('pre').sort_index(ascending=False)  # 按照预测概率从大到小排序作为阈值
        crossdens = df_data.cumsum(axis=0) / df_data.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        auc = np.trapz(y=tpr, x=fpr)
        return auc, crossdens

    @classmethod
    def get_ks(cls, df_pre, df_label, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签
        :return:
            'ks': KS值
            'crossdens': 好坏客户累积概率分布以及其差值gap
        '''
        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        crossfreq = pd.crosstab(df_pre, df_label).sort_index(ascending=False)  # 按概率降序排序
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        crossdens['gap'] = abs(crossdens['fpr'] - crossdens['tpr'])
        ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
        return ks, crossdens

    @classmethod
    def get_g7_ks(cls, df_pre, df_got_fee, df_report_fee):
        '''
        功能: 计算auc值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_got_fee: 一维数组或series，代表已赚保费
        :param df_report_fee: 一维数组或series，代表赔付额
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        df_data = pd.concat(
            [df_pre.reset_index(drop=True), df_got_fee.reset_index(drop=True), df_report_fee.reset_index(drop=True)],
            axis=1)
        df_data.columns = ['pre', 'got_fee', 'report_fee']
        df_data = df_data.set_index('pre').sort_index(ascending=False)  # 按照预测概率从大到小排序作为阈值
        crossdens = df_data.cumsum(axis=0) / df_data.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        crossdens['gap'] = abs(crossdens['fpr'] - crossdens['tpr'])
        ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
        return ks, crossdens

# 评分校准
def get_calibration_score(raw_score):
    pass

if __name__ == '__main__':
    model_score = ScoreCardModel()
    score = model_score.probability_to_score(0.5)
    print(score)